      program pwe2d
c
c     ******************************************************************
c
c     Calculates photonic structure of two-dimensional photonic crystals
c     by plane-wave expansion.
c     Written by L.C. Andreani, Department of Physics "A. Volta",
c     University of Pavia, Italy. e-mail andreani@fisicavolta.unipv.it
c
c     ******************************************************************
c
      implicit none
      integer jlattice,jbasis,j,n1,n2,maxg,maxpw,npw,maxham,nrec
      integer nk,nktot,jk,maxk,j1,j2,jparxy,nham,ifail,jband,jeigenv
      integer nkh,job,nbands,maxbands,jparkz,maxnkz
      integer jkeigenv,jstate,ngrid,jxgrid,jygrid,jcomp,n1max,n2max
      integer jfreephotons,nvar,jloop,jvar,ngap,jgap,nham_store
      real*8 eps1,eps2,pi,g,upper,fill,factor
      real*8 alength1,alength2,alength3,alength4
      real*8 ft11,ft12,ft13,ft98,ft99,ft22,ft40,ft41
      real*8 ftcheck,rcond,dummy,bl1,bl2,supercellperiod
      real*8 a1(2),a2(2),aret,ucarea,b1(2),b2(2),bzarea
      real*8 xmin,xmax,ymin,ymax,delxgrid,delygrid,x,y,emod,bmod
      real*8 varmin,varmax,var,delvar,fill_loop
      complex*16 i
      parameter (aret=1d0,maxg=2e5,maxk=1000,maxpw=1000,maxham=2*maxpw)
      parameter (maxbands=100,upper=1d-12,i=(0d0,1d0),maxnkz=200)
      integer n1vec(maxg),n2vec(maxg),iwksp(maxg),nwksp(maxg),nkz,jkz
      integer jwork(maxham)
      real*8 alair(maxk),akvec(maxk)
      real*8 gx,gy,gmod(maxg),gxvec(maxg),gyvec(maxg),wksp(maxg)
      real*8 akxvec(maxk),akyvec(maxk),ax(maxk),gz1,gz2
      real*8 akx,aky,gx1,gy1,gx2,gy2,akz,akzmin,akzmax,delkz,refrindex
      real*8 h(maxham,maxham),energ(maxham),energ2(maxham)
      real*8 w1(maxham),w2(maxham),eigenv(maxham,maxham),axmult,axsum
      real*8 u(maxham,maxham),y0
      real*8 square_g(2),square_x(2),square_m(2)
      real*8 triang_g(2),triang_m(2),triang_k(2)
      real*8 rect_g(2),rect_x(2),rect_z(2),rect_l(2),rect_m(2),rect_y(2)
      real*8 w1_g(2),w1_x1(2),w1_x2(2),w1_m(2)
      real*8 gmod1,alpha1,beta1,gamma1,denom1,pol1(3),qol1(3)
      real*8 gmod2,alpha2,beta2,gamma2,denom2,pol2(3),qol2(3)
      real*8 bands(maxk,maxbands),ecoeff(maxpw,3),bcoeff(maxpw,3)
      real*8 delta_kronecker,edge(2*maxbands)
      real*8 emin(maxbands),emax(maxbands),diff(maxbands)
      real*8 akzvec(0:maxnkz),bands_store(maxk,maxbands,0:maxnkz)
      complex*16 efield(3),bfield(3)
      real*8 ft(maxpw,maxpw),ftinv(maxpw,maxpw)
      character*4 dir
      character*12 filename
      common/dirlattice/a1,a2,ucarea
      common/reclattice/b1,b2,bzarea
      common/eps/eps1,eps2,fill
      common/length/alength1,alength2,alength3,alength4
      common/ftinv/ftinv,gxvec,gyvec
      common/iarray/jwork
      common/matrix/u

      open (unit=1,file='pwe2d.dat',status='old')
      open (unit=2,file='pwe2d.out',status='unknown')
      open (unit=3,file='pwe2d_gvec.out',status='unknown')
      open (unit=7,file='pwe2d_vl.out',status='unknown')
      open (unit=9,file='pwe2d_refrindex.out',status='unknown')
      open (unit=28,file='pwe2d_gapwidth.out',status='unknown')
      open (unit=29,file='pwe2d_gapedge.out',status='unknown')
      open (unit=30,file='pwe2d_offplane.out',status='unknown')
      open (unit=31,file='pwe2d_ll.out',status='unknown')

      pi=4d0*atan(1d0)
      read (1,*) jlattice
      read (1,*) jbasis,alength1,alength2,alength3,alength4
      read (1,*) eps1,eps2
      read (1,*) jparxy,jparkz
      read (1,*) akzmin,akzmax,nkz
      read (1,*) nk
      read (1,*) npw
      read (1,*) jeigenv,jkeigenv,jstate
      read (1,*) factor
      read (1,*) xmin,xmax,ymin,ymax,ngrid
      read (1,*) jfreephotons
      read (1,*) jloop,varmin,varmax,nvar
      close (1)
c
c     checks values of jparxy and dimension of hamiltonian
c
      if (jparxy.eq.0.or.jparxy.eq.1) then
        nham=npw
      else if (jparxy.eq.-1) then
        nham=2*npw
        if (jparkz.eq.0.or.jparkz.eq.1) then
         stop ' vertical parity symmetry not implemented if jparxy=-1'
        endif
      else 
        stop ' jparxy must be -1,0,1'
      endif
      nham_store=nham
      if (npw.gt.maxpw) stop ' npw>maxpw'
      if (nham.gt.maxham) stop ' nham>maxham'
      if (jeigenv.lt.0.and.jeigenv.gt.1) stop ' jeigenv must be 0 or 1'
      if (nkz.gt.maxnkz) stop ' nkz>maxnkz'


c
c     Calculates lattice constant when a supercell is used,
c     in the case of a 1D lattice with microcavity.
c
      supercellperiod=aret
      if (jlattice.eq.1.and.jbasis.eq.98) then
        n1=int(alength3)
        if (n1.ne.alength3) stop ' supercell: alength3 must be integer'
        n2=n1-1
        supercellperiod=alength2+n1*alength1+n2*(1d0-alength1)
        print *, 'supercellperiod=',supercellperiod
        factor=factor/supercellperiod
      endif
        
c
c     defines primitive vectors of (direct) Bravais lattice
c     and unit cell area
c
      if (jlattice.eq.1) then                ! square lattice
        a1(1)=aret
        a1(2)=0d0
        a2(1)=0d0
        a2(2)=aret
      else if (jlattice.eq.2) then           ! triangular lattice
        a1(1)=aret
        a1(2)=0d0
        a2(1)=aret/2d0
        a2(2)=sqrt(3d0)/2d0*aret
      else if (jlattice.eq.4) then           ! W1 waveguide
        a1(1)=aret
        a1(2)=0d0
        a2(1)=0d0
        a2(2)=aret*(alength2+alength3)*sqrt(3d0)
      else
        stop ' jlattice must be 1,2 or 4'
      endif
      ucarea=abs(a1(1)*a2(2)-a1(2)*a2(1))
c
c     defines primitive vectors of reciprocal lattice and unit cell area
c
      b1(1)= 2d0*pi/ucarea*a2(2)
      b1(2)=-2d0*pi/ucarea*a2(1)
      b2(1)=-2d0*pi/ucarea*a1(2)
      b2(2)= 2d0*pi/ucarea*a1(1)
      bzarea=4d0*pi*pi/ucarea

c
c     calculates reciprocal lattice vectors
c

      bl1=sqrt(b1(1)*b1(1)+b1(2)*b1(2))
      bl2=sqrt(b2(1)*b2(1)+b2(2)*b2(2))
      if (bl1.le.bl2) then
        n2max=50
        n1max=n2max*bl2/bl1
      else
        n1max=50
        n2max=n1max*bl1/bl2
      endif
      if (jlattice.eq.1.and.jbasis.eq.98.or.jbasis.eq.99) n2max=0

      j=0
!      print *, 'WARNING: WHEN JBASIS=99 ONLY REC. LATT. VECTORS ALONG X'
      do n1=-n1max,n1max
      do n2=-n2max,n2max
        j=j+1
        if (j.gt.maxg) stop ' j>maxg'
        gx=n1*b1(1)+n2*b2(1)
        gy=n1*b1(2)+n2*b2(2)
        gmod(j)=sqrt(gx*gx+gy*gy)
        n1vec(j)=n1
        n2vec(j)=n2
        gxvec(j)=gx
        gyvec(j)=gy
      enddo
      enddo
      nrec=j
      call sort5 (nrec,gmod,gxvec,gyvec,n1vec,n2vec,wksp,iwksp,nwksp)
      do j=1,10*npw
        write (3,12) j,n1vec(j),n2vec(j),gxvec(j),gyvec(j),gmod(j)
      enddo
12    format (1x,i4,4x,i3,2x,i3,4x,f10.4,2x,f10.4,4x,f14.4)


c
c     loop over variable VAR (dielectric constant or hole radius or...)
c
      if (jloop.eq.1) then
        write (28,24) jlattice,jbasis
        write (28,26) eps1,eps2
        write (28,27) jparxy,jparkz
        write (28,28) npw,factor
        write (29,24) jlattice,jbasis
        write (29,26) eps1,eps2
        write (29,27) jparxy,jparkz
        write (29,28) npw,factor
!        if (jlattice.eq.1) then          ! loop on filling factor
!          varmin=pi*varmin*varmin
!          varmax=pi*varmax*varmax
!        else if (jlattice.eq.2) then     ! loop on filling factor
!          varmin=2d0*pi/sqrt(3d0)*varmin*varmin
!          varmax=2d0*pi/sqrt(3d0)*varmax*varmax
!        endif
      endif
      if (nvar.ne.0) delvar=(varmax-varmin)/float(nvar)
      if (jloop.eq.0.and.nvar.ne.0) stop 'no loop over VAR if JLOOP=0'
      do 9995 jvar=0,nvar
      var=varmin+jvar*delvar
      if (jloop.eq.1) then
         print *, 'loop variable is alength1'
         alength1=var
!         fill_loop=var  ! used fo rloop over filling factor
!       alength1=sqrt(fill_loop/pi)    ! square lattice
!       alength1=sqrt(sqrt(3d0)*fill_loop/2d0/pi)    ! tiangular lattice

         if (jvar+1.lt.10) write (filename,9991) jvar+1
         if (jvar+1.ge.10) write (filename,9992) jvar+1
9991     format('pwe2d00',i1,'.out')
9992     format('pwe2d0',i2,'.out')
         close (2)
!         close (9)
         print *, 'writing output on file ',filename
         open (unit=2,file=filename,status='unknown')
      endif


c
c     calculates matrix with Fourier transform of dielectric constant
c
      do 120 j1=1,npw
      do 120 j2=1,npw
        gx1=gxvec(j1)
        gy1=gyvec(j1)
        gx2=gxvec(j2)
        gy2=gyvec(j2)
c
c       Fourier transform of dielectric constant
c
        if (jlattice.eq.1) then       ! square lattice
          if (jbasis.eq.1) then           ! circular rods
            ft(j1,j2)=ft11(gx1-gx2,gy1-gy2)
          else if (jbasis.eq.2) then      ! square rods, non rotated
            ft(j1,j2)=ft12(gx1-gx2,gy1-gy2)
          else if (jbasis.eq.3) then      ! square rods, 45 rotated
            ft(j1,j2)=ft13(gx1-gx2,gy1-gy2)
          else if (jbasis.eq.98) then     ! 1D stripes with microcavity
            ft(j1,j2)=ft98(gx1-gx2,gy1-gy2)
          else if (jbasis.eq.99) then     ! 1D stripes
            ft(j1,j2)=ft99(gx1-gx2,gy1-gy2)
          else
            stop' when jlattice=1 jbasis must be 1,2,3, 98 or 99'
          endif
        else if (jlattice.eq.2) then
          if (jbasis.eq.1) then           ! circular rods, triangular
            ft(j1,j2)=ft11(gx1-gx2,gy1-gy2)
          else if (jbasis.eq.2) then      ! circular rods, graphite
            ft(j1,j2)=ft22(gx1-gx2,gy1-gy2)
          else
            stop ' when jlattice=2 jbasis must be 1 or 2'
          endif
        else if (jlattice.eq.4) then
          if (jbasis.eq.0) then     ! bulk with W1 supercell, my routine
            ft(j1,j2)=ft40(gx1-gx2,gy1-gy2)
          else if (jbasis.eq.1) then      ! W1 waveguide, routine Agio
            ft(j1,j2)=ft41(gx1-gx2,gy1-gy2)
          else
            stop ' when jlattice=4 jbasis must be 0 or 1'
          endif
        endif
        if (jfreephotons.eq.1) then
          ft(j1,j2)=delta_kronecker(j1,j2)*ft(1,1)
        endif
120   continue
      write (6,121) fill
121   format (1x,'filling fraction=',f12.6)
c
c     Fourier transform of inverse dielectric constant (Ho's method)
c     Linpack routines are used (http://www.netlib.org/linpack)
c
      print *, ' starts inversion of dielectric matrix'
      call savemt (ft,ftinv,maxpw,npw)
      call dpoco (ftinv,maxpw,npw,rcond,w1,ifail)
      if (ifail.gt.0) stop ' after DPOCO: ifail>0'
      if (abs(rcond).lt.1d-14) stop ' after DPOCO: rcond too small'
      job=1
      call dpodi (ftinv,maxpw,npw,dummy,job)
      do 122 j1=2,npw
       do 122 j2=1,j1-1
122      ftinv(j1,j2)=ftinv(j2,j1)
      call checkinv (ft,ftinv,maxpw,npw)
      print *, ' ends inversion of dielectric matrix'
c
c     prepares k-vectors for calculating photonic bands (akxvec,akyvec)
c     and x-coordinate vector for plotting (ax)
c
      axmult=1d0
      axsum=0d0
      if (jlattice.eq.1) then
        call square_points (square_g,square_x,square_m)
        if (jbasis.eq.99) then
         nktot=nk+1  
         dir(1:4)='xgam'
         call kline (square_g,square_x,akxvec,akyvec,ax,maxk,nk,1)
        else

!         nktot=nk+1
!         call kline (square_x,square_g,akxvec,akyvec,ax,maxk,nk,1)
!         dir(1:4)='xgam'
!         axmult=1d0/(1d0+sqrt(2d0))
!         axsum=0d0

!         nktot=nk+1
!         call kline (square_g,square_m,akxvec,akyvec,ax,maxk,nk,1)
!         dir(1:4)='gamm'
!         axmult=sqrt(2d0)/(1d0+sqrt(2d0))
!         axsum=1d0/(1d0+sqrt(2d0))

!         nkh=nk/2
!         nktot=2*nk+nkh+1
!         call kline (square_x,square_g,akxvec,akyvec,ax,maxk,nk,1)
!         call kline (square_g,square_m,akxvec,akyvec,ax,maxk,nk,1+nk)
!         call kline (square_m,square_x,akxvec,akyvec,ax,maxk,nkh,1+2*nk)

         nktot=3*nk+1
         call kline (square_g,square_x,akxvec,akyvec,ax,maxk,nk,1)
         call kline (square_x,square_m,akxvec,akyvec,ax,maxk,nk,1+nk)
         call kline (square_m,square_g,akxvec,akyvec,ax,maxk,nk,1+2*nk)

        endif
      else if (jlattice.eq.2) then
        call triang_points (triang_g,triang_m,triang_k)

!        nktot=nk+1
!        call kline (triang_m,triang_g,akxvec,akyvec,ax,maxk,nk,1)
!        dir(1:4)='mgam'
!        axmult=1d0/(1d0+2d0/sqrt(3d0))
!        axsum=0d0

!        nktot=nk+1
!        call kline (triang_g,triang_k,akxvec,akyvec,ax,maxk,nk,1)
!        dir (1:4)='gamk'
!        axmult=2d0/sqrt(3d0)/(1d0+2d0/sqrt(3d0))
!        axsum=1d0/(1d0+2d0/sqrt(3d0))

!        nktot=nk+1
!        call kline (triang_k,triang_m,akxvec,akyvec,ax,maxk,nk,1)
!        dir (1:4)='kapm'
!        axmult=1d0
!        axsum=0d0

!        nktot=2*nk+1
!        call kline (triang_m,triang_g,akxvec,akyvec,ax,maxk,nk,1)
!        call kline (triang_g,triang_k,akxvec,akyvec,ax,maxk,nk,1+nk)

!        nkh=nk/2
!        nktot=2*nk+nkh+1
!        call kline (triang_m,triang_g,akxvec,akyvec,ax,maxk,nk,1)
!        call kline (triang_g,triang_k,akxvec,akyvec,ax,maxk,nk,1+nk)
!        call kline (triang_k,triang_m,akxvec,akyvec,ax,maxk,nkh,1+2*nk)

        nktot=3*nk+1
        call kline (triang_g,triang_m,akxvec,akyvec,ax,maxk,nk,1)
        call kline (triang_m,triang_k,akxvec,akyvec,ax,maxk,nk,1+nk)
        call kline (triang_k,triang_g,akxvec,akyvec,ax,maxk,nk,1+2*nk)

      else if (jlattice.eq.4) then
!        call square_points (square_g,square_x,square_m) 
!        call kline (square_g,square_x,akxvec,akyvec,ax,maxk,nk,1)

        call w1_points (w1_g,w1_x1,w1_x2,w1_m)

        nktot=nk+1
        dir(1:4)='xgam'
        call kline (w1_g,w1_x1,akxvec,akyvec,ax,maxk,nk,1)

!        nktot=2*nk+1
!        call kline (w1_x2,w1_g,akxvec,akyvec,ax,maxk,nk,1)
!        call kline (w1_g,w1_x1,akxvec,akyvec,ax,maxk,nk,1+nk)


!        nktot=4*nk+1
!        call kline (w1_g,w1_x1,akxvec,akyvec,ax,maxk,nk,1)
!        call kline (w1_x1,w1_m,akxvec,akyvec,ax,maxk,nk,1+nk)
!        call kline (w1_m,w1_x2,akxvec,akyvec,ax,maxk,nk,1+2*nk)
!        call kline (w1_x2,w1_g,akxvec,akyvec,ax,maxk,nk,1+3*nk)


      endif

c
c     normalizes ax vector in order that its values are between 0 and 1
c
      if (nktot.gt.maxk) stop ' nktot>maxk'
      do jk=1,nktot
        if (abs(akxvec(jk)).lt.upper) akxvec(jk)=akxvec(jk)+upper
        if (abs(akyvec(jk)).lt.upper) akyvec(jk)=akyvec(jk)+upper
        if (nk.ne.0) ax(jk)=ax(jk)/ax(nktot)*axmult+axsum
      enddo

22    format (1x,i4,4x,3(f12.6,2x))
c
c     starts loop over wavevectors
c
      if (nkz.ne.0) delkz=(akzmax-akzmin)/nkz
      do 110 jkz=0,nkz
      akz=akzmin+jkz*delkz
      akzvec(jkz)=akz
      do 100 jk=1,nktot
        akx=akxvec(jk)
        aky=akyvec(jk)
        akvec(jk)=sqrt(akx*akx+aky*aky)
        alair(jk)=akvec(jk)/2d0/pi*factor
        call defham (akx,aky,akz,jparxy,h,npw)
        if ((jparxy.eq.0.or.jparxy.eq.1).and.
     &      (jparkz.eq.0.or.jparkz.eq.1))
     &    call decouple(akx,aky,npw,h,jlattice,jparxy,jparkz,dir,nham)
        write (6,101) jk,akx,aky,nham
101     format (1x,i3,2x,f12.6,2x,f12.6,2x,i4)
        if (nham.eq.0) goto 100
        call checksym (h,maxham,nham)
!        write (6,*), 'jk=',jk
!        wrmat (h,maxham,nham,1d0,0d0)
!        if (jparxy.eq.0.or.jparxy.eq.1) call wrmat (h,maxham,nham,1d0,0d0)
!        if (jparxy.eq.-1) then
!          call deblock (h,npw,eigenv,2)
!          call wrmat (eigenv,maxham,npw,1d0,0d0)
!        endif
c
c       solves eigenvalue problem
c
!        print *, ' starts diagonalization of hamiltonian'
        ifail=0
        call rs (maxham,nham,h,energ2,jeigenv,eigenv,w1,w2,ifail)

c
c       restores eigenvectors in original basis, if parity symmetry was used
c
        if (jeigenv.eq.1.and.(jparkz.eq.0.or.jparkz.eq.1))
     &    call restore (eigenv,nham,nham_store)

!          call wrmat (eigenv,maxham,nham,1d0,0d0)


!        call wrvec (energ2,maxham,nham,1d0,0d0)
!        print *, ' ends diagonalization of hamiltonian'
c
c       checks energies are positive, rescales in dimensionless units
c
        do j1=1,nham
          if (energ2(j1).ge.0) then
            energ(j1)=sqrt(energ2(j1))*aret/2d0/pi*factor
          else if (-1d-8.lt.energ2(j1)) then
            energ(j1)=0d0
          else
            energ(j1)=-99.
          endif
        enddo
        do j1=1,maxbands
          bands(jk,j1)=energ(j1)
        enddo
c
c       writes results
c
        write (31,180) ax(jk),alair(jk)
!        write (31,99) ax(jk),akx/2d0/pi,aky/2d0/pi,
!     &               (energ(jband),jband=1,10)
!        write (32,99) ax(jk),akx/2d0/pi,aky/2d0/pi,
!     &               (energ(jband),jband=11,20)
!        write (33,99) ax(jk),akx/2d0/pi,aky/2d0/pi,
!     &               (energ(jband),jband=21,30)
99      format (1x,f8.4,1x,2(f6.2,1x),12(1x,f7.4))


c
c       Starts calculation of electric and magnetic fields.
c       Method: finds vector coefficients ECOEFF, BCOEFF of the 
c       plane waves, then calculates fields in a square grid.
c
        if (jeigenv.eq.1.and.jk.eq.jkeigenv) then
          do 210 j1=1,npw
           do jcomp=1,3
             ecoeff(j1,jcomp)=0d0
             bcoeff(j1,jcomp)=0d0
           enddo
           gx1=akx+gxvec(j1)
           gy1=aky+gyvec(j1)
           gz1=akz
           gmod1=sqrt(gx1*gx1+gy1*gy1+gz1*gz1)
           alpha1=gx1/gmod1
           beta1 =gy1/gmod1
           gamma1=gz1/gmod1
           denom1=sqrt(alpha1*alpha1+beta1*beta1)
           pol1(1)=gamma1*alpha1/denom1
           pol1(2)=gamma1*beta1 /denom1
           pol1(3)=-denom1
           qol1(1)=-beta1/denom1
           qol1(2)=alpha1/denom1
           qol1(3)=0d0
           if (jparxy.eq.1) then 
             do jcomp=1,3
               bcoeff(j1,jcomp)=qol1(jcomp)*eigenv(j1,jstate)
             enddo
           else if (jparxy.eq.0) then 
             bcoeff(j1,3)=eigenv(j1,jstate)
           else if (jparxy.eq.-1) then 
             do jcomp=1,3
               bcoeff(j1,jcomp)=pol1(jcomp)*eigenv(2*j1-1,jstate)
     &                         +qol1(jcomp)*eigenv(2*j1  ,jstate)
             enddo
           endif
           do 215 j2=1,npw
             gx2=akx+gxvec(j2)
             gy2=aky+gyvec(j2)
             gz2=akz
             gmod2=sqrt(gx2*gx2+gy2*gy2+gz2*gz2)
             alpha2=gx2/gmod2
             beta2 =gy2/gmod2   
             gamma2=gz2/gmod2
             denom2=sqrt(alpha2*alpha2+beta2*beta2)
             pol2(1)=gamma2*alpha2/denom2
             pol2(2)=gamma2*beta2 /denom2
             pol2(3)=-denom2  
             qol2(1)=-beta2/denom2
             qol2(2)=alpha2/denom2
             qol2(3)=0d0
             if (jparxy.eq.1) then 
               ecoeff(j1,3)=ecoeff(j1,3)
     &         -aret/2d0/pi/energ(jstate)*gmod2*
     &          ftinv(j1,j2)*eigenv(j2,jstate)
             else if (jparxy.eq.0) then 
               do jcomp=1,3
                 ecoeff(j1,jcomp)=ecoeff(j1,jcomp)
     &           +aret/2d0/pi/energ(jstate)*factor*gmod2
     &           *ftinv(j1,j2)*qol2(jcomp)*eigenv(j2,jstate)
               enddo
             else if (jparxy.eq.-1) then 
               do jcomp=1,3
                ecoeff(j1,jcomp)=ecoeff(j1,jcomp)-aret/2d0/pi
     &            /energ(jstate)*factor*gmod2*ftinv(j1,j2)*
     &             ( qol2(jcomp)*eigenv(2*j2-1,jstate)
     &              -pol2(jcomp)*eigenv(2*j2  ,jstate) )
               enddo
             endif
215        continue
210       continue


          if (ngrid.ne.0) delxgrid=(xmax-xmin)/float(ngrid)
          if (ngrid.ne.0) delygrid=(ymax-ymin)/float(ngrid)
          do 230 jxgrid=0,ngrid
           do 231 jygrid=0,ngrid
            x=xmin+delxgrid*jxgrid
            y=ymin+delygrid*jygrid
            x=x/supercellperiod
            y=y/supercellperiod
            emod=0d0
            bmod=0d0
            do 235 jcomp=1,3
              efield(jcomp)=(0d0,0d0)
              bfield(jcomp)=(0d0,0d0)
              do 240 j1=1,npw
                efield(jcomp)=efield(jcomp)+ecoeff(j1,jcomp)*
     &           exp(i*( (akx+gxvec(j1))*x+(aky+gyvec(j1))*y ))
                bfield(jcomp)=bfield(jcomp)+bcoeff(j1,jcomp)*
     &           exp(i*( (akx+gxvec(j1))*x+(aky+gyvec(j1))*y ))
240           continue
              emod=emod+efield(jcomp)*dconjg(efield(jcomp))
              bmod=bmod+bfield(jcomp)*dconjg(bfield(jcomp))
235         continue
            if (jparxy.eq.1) then
              write (15,242) x,y,dble(efield(3))
              write (16,242) x,y,dimag(efield(3))
              write (17,242) x,y,(bfield(jcomp),jcomp=1,2)
            else if (jparxy.eq.0) then
              write (15,242) x,y,dble(bfield(3))
              write (16,242) x,y,dimag(bfield(3))
              write (17,242) x,y,(efield(jcomp),jcomp=1,2)
            else if (jparxy.eq.-1) then
              write (15,242) x,y,(efield(jcomp),jcomp=1,3)
              write (16,242) x,y,(bfield(jcomp),jcomp=1,3)
            endif
            write (18,243) x,y,emod
            write (19,243) x,y,bmod
231        continue
           write (15,*)
           write (16,*)
           write (17,*)
           write (18,*)
           write (19,*)
230       continue
242       format(2(1x,f8.4),1x,3(1x,e9.3,' ',e9.3))
243       format(2(1x,f8.4),1x,2(1x,e10.4))
        endif

c
c       ends loop over wavevectors
c
100   continue
      print *, ' NORMALIZATION FACTOR FOR BANDS=',factor
c
c     writes output for gnuplot
c
      nbands=min(maxbands,nham)
      write (6,23) nktot,nbands
23    format(1x,'number of wavevectors=',i3,/,' number of bands=',i4)

      write (2,24) jlattice,jbasis
      write (2,25) alength1,alength2,alength3
      write (2,26) eps1,eps2
      write (2,27) jparxy,jparkz
      write (2,28) npw,factor
24    format('# jlattice=',i2,'  jbasis=',i2)
25    format('# alength1=',f10.4,' alength2=',f10.4,' alength3=',f10.4)
26    format('# eps1=',f10.4,'     eps2=',f10.4)
27    format('# jparxy=',i2,'  jparkz=',i2)
28    format('# npw=',i3,'  factor=',f10.4)


      do j1=1,nbands
        write (2,179) j1
        do jk=1,nktot
          write (2,180) ax(jk),bands(jk,j1)
        enddo
        write (2,*)
      enddo
179   format('#   band number=',i3)
180   format (1x,f12.6,3(2x,e12.6))
      if (bands(1,1).ne.0) refrindex=akz*factor/bands(1,1)/2d0/pi
      write (9,181) akz,(bands(1,j1),j1=1,10),refrindex
181   format (1x,f12.6,10(2x,e10.4),2x,f12.6)

         do j1=1,nbands
          do jk=1,nktot
            bands_store(jk,j1,jkz)=bands(jk,j1)
          end do
         end do

110   continue                        ! ends loop over kz
      open (unit=7,file='pwe2d_vl.out',status='unknown')
      if (nk.ne.0) then
        write (*,182) (ax(jk),jk=1,nktot,nk)
        do jk=nk+1,nktot-2,nk
          write (7,180) ax(jk),0d0
          write (7,180) ax(jk),10d0
          write (7,180) 
        enddo
      endif
182   format (' set xtics (',6(f8.5,','),f8.5,')')
      close (7)

      if (nkz.ne.0) then
       do j1=1,20
        do jkz=0,nkz
         write (30,180) 
     &                  akzvec(jkz),
!     &                  akzvec(jkz)*alength2/aret,
     &                  bands_store(jkeigenv,j1,jkz)
        end do
        write (30,*)
       end do

       do jkz=0,nkz
!         write (31,180) akzvec(jkz),akzvec(jkz)*factor/2d0/pi ! air light line
!         write (31,180) akzvec(jkz)*alength2/aret,
!     &          akzvec(jkz)*factor/2d0/pi,
!     &          akzvec(jkz)*factor/2d0/pi/sqrt(eps1),
!     &          akzvec(jkz)*factor/2d0/pi/sqrt(eps2)
       end do
!       write (31,*)
!       do jkz=0,nkz
!         write (31,180) akzvec(jkz),akzvec(jkz)*factor/2d0/pi ! air light line
!         write (31,180) akzvec(jkz),
!     &          akzvec(jkz)*factor/2d0/pi/sqrt(eps2)
!       end do

      end if
      close (30)
      close (31)

c
c     calculates gap map
c
      call extrema (bands,maxk,maxbands,nktot,nbands,emin,emax)
      ngap=0
      do j1=1,nbands-1
        diff(j1)=emin(j1+1)-emax(j1)
        if (diff(j1).gt.0.00001.and.emax(j1).lt.1.0) then
          ngap=ngap+1
          edge(2*ngap-1)=emax(j1)
          edge(2*ngap  )=emin(j1+1)
        endif
      enddo
      write (28,184) alength1,fill,(diff(j1),j1=1,12)
      write (29,185) alength1,fill,(edge(jgap),jgap=1,2*ngap)
184   format(1x,f9.5,1x,f9.5,1x,12(1x,f7.4))
185   format(1x,f9.5,1x,f9.5,1x,20(1x,f7.4))

c
c     ends loop over filling factor
c
9995  continue

      close (2)
      close (3)
      close (7)
      stop
      end

c
      subroutine defham (akx,aky,akz,jparxy,h,npw)
c
c     defines hamiltonian matrix
c
      implicit none
      integer jparxy,maxpw,maxham,npw,maxg,j1,j2
      parameter (maxpw=1000,maxham=2*maxpw,maxg=2e5)
      real*8 akx,aky,akz,gx1,gy1,gz1,gx2,gy2,gz2,scal
      real*8 h(maxham,maxham),ftinv(maxpw,maxpw)
      real*8 gxvec(maxg),gyvec(maxg)
      real*8 gmod1,denom1,alpha1,beta1,gamma1,pol1(3),qol1(3)
      real*8 gmod2,denom2,alpha2,beta2,gamma2,pol2(3),qol2(3)
      common/ftinv/ftinv,gxvec,gyvec
      do 5 j1=1,npw
      do 5 j2=1,npw
        gx1=akx+gxvec(j1)
        gy1=aky+gyvec(j1)
        gx2=akx+gxvec(j2)
        gy2=aky+gyvec(j2)
        gz1=akz
        gz2=akz
        if (jparxy.eq.1) then        ! E-polarization
          gmod1=gx1*gx1+gy1*gy1
          gmod2=gx2*gx2+gy2*gy2
          h(j1,j2)=sqrt(gmod1)*sqrt(gmod2)*ftinv(j1,j2)
        else if (jparxy.eq.0) then   ! H-polarization
          h(j1,j2)=(gx1*gx2+gy1*gy2)*ftinv(j1,j2)
        else if (jparxy.eq.-1) then   ! both polarizations together
          gmod1=sqrt(gx1*gx1+gy1*gy1+gz1*gz1)
          gmod2=sqrt(gx2*gx2+gy2*gy2+gz2*gz2)
c 
c         director cosines
c
          alpha1=gx1/gmod1
          beta1 =gy1/gmod1
          gamma1=gz1/gmod1
          denom1=sqrt(alpha1*alpha1+beta1*beta1)
c
          alpha2=gx2/gmod2
          beta2 =gy2/gmod2
          gamma2=gz2/gmod2
          denom2=sqrt(alpha2*alpha2+beta2*beta2)
c
c         polarization vectors
c
          pol1(1)=gamma1*alpha1/denom1
          pol1(2)=gamma1*beta1 /denom1
          pol1(3)=-denom1
          qol1(1)=-beta1/denom1
          qol1(2)=alpha1/denom1
          qol1(3)=0d0
c
          pol2(1)=gamma2*alpha2/denom2
          pol2(2)=gamma2*beta2 /denom2
          pol2(3)=-denom2
          qol2(1)=-beta2/denom2
          qol2(2)=alpha2/denom2
          qol2(3)=0d0
c
c         elements of hamiltonian matrix
c
          h(2*j1-1,2*j2-1)=scal(qol1,qol2,3)*gmod1*gmod2*ftinv(j1,j2)
          h(2*j1-1,2*j2  )=scal(qol1,pol2,3)*gmod1*gmod2*ftinv(j1,j2)
          h(2*j1  ,2*j2-1)=scal(pol1,qol2,3)*gmod1*gmod2*ftinv(j1,j2)
          h(2*j1  ,2*j2  )=scal(pol1,pol2,3)*gmod1*gmod2*ftinv(j1,j2)
        endif
5     continue
      return
      end
c
      subroutine extrema (bands,maxk,maxbands,nktot,nbands,emin,emax)
c
c     finds min and max of each band for gap map
c
      implicit none
      integer maxk,maxbands,nktot,nbands,jb,jk
      real*8 bands(maxk,maxbands),emin(maxbands),emax(maxbands)
      real*8 diff(maxbands),amin,amax
      if (nktot.gt.maxk) stop ' from EXTREMA: nktot>maxk'
      if (nbands.gt.maxbands) stop ' from EXTREMA: nbands>maxbands'
      do jb=1,nbands
        amin=1d10
        amax=0d0
        do jk=1,nktot
          amin=min(amin,bands(jk,jb))
          amax=max(amax,bands(jk,jb))
        enddo
        emin(jb)=amin
        emax(jb)=amax
      enddo
      return
      end
c
      subroutine decouple(akx,aky,npw,h,jlattice,jparxy,jparkz,dir,nham)
c
c     Decouples hamiltonian matrix in blocks of even and odd parities
c     w.r.t. a specified mirror plane determined by string "dir".
c
      implicit none
      integer jlattice,jparxy,jparkz,maxpw,maxham,npw,maxg,j1,j2,j3
      integer neven,nodd,nham,jeven,jodd
      parameter (maxpw=1000,maxham=2*maxpw,maxg=2e5)
      real*8 h(maxham,maxham),ftinv(maxpw,maxpw),u(maxham,maxham)
      real*8 work(maxpw,maxpw)
      real*8 akx,aky,gxvec(maxg),gyvec(maxg),upper,gx1sym,gy1sym,gx2,gy2
      real*8 cos45,pi,gx1,gy1
      character*4 dir
      integer jsym(maxpw),jwork(maxham)
      parameter (upper=1d-12)
      common/ftinv/ftinv,gxvec,gyvec
      common/iarray/jwork
      common/matrix/u
      pi=4d0*atan(1d0)
c
c     for each reciprocal lattice vector, finds the symmetric one
c
      do 10 j1=1,npw
        gx1=gxvec(j1)
        gy1=gyvec(j1)
        if (jlattice.eq.1) then
          if (dir(1:4).eq.'xgam') then
            gx1sym= gx1
            gy1sym=-gy1
          else if (dir(1:4).eq.'gamm') then
            gx1sym=gy1
            gy1sym=gx1
          else
            stop ' from DECOUPLE: wrong value of string DIR'
          endif
        else if (jlattice.eq.2) then
          if (dir(1:4).eq.'mgam') then
            gx1sym=-gx1
            gy1sym= gy1
          else if (dir(1:4).eq.'gamk') then
            gx1sym=(-gx1+sqrt(3d0)*gy1)/2d0
            gy1sym=(sqrt(3d0)*gx1+gy1)/2d0
          else if (dir(1:4).eq.'kapm') then
            gx1sym=gx1
            gy1sym=-gy1
          else
            stop ' from DECOUPLE: wrong value of string DIR'
          endif
        else if (jlattice.eq.4) then
          if (dir(1:4).eq.'xgam') then
            gx1sym= gx1
            gy1sym=-gy1
          else
            stop ' from DECOUPLE: wrong value of string DIR'
          endif
        else
          stop ' DECOUPLE implemented only for jlattice=1,2,4'
        endif
        jsym(j1)=0
        do 20 j2=1,npw
          gx2=gxvec(j2)
          gy2=gyvec(j2)
          if (abs(gx2-gx1sym).le.upper.and.abs(gy2-gy1sym).lt.upper) 
     &      jsym(j1)=j2
20      continue
        if (jsym(j1).eq.0) stop ' from DECOUPLE: no sym. wavevector'
!        write (99,999) j1,gxvec(j1),gyvec(j1),
!     &                 jsym(j1),gxvec(jsym(j1)),gyvec(jsym(j1))
10    continue
999   format (1x,3(i3,2x,f12.6,2x,f12.6,2x,2x))
c
c     calculates the unitary matrix which decouples the hamiltonian
c     in the basis of even and odd states
c
      cos45=sqrt(0.5d0)
      do 30 j1=1,npw
      do 30 j2=1,npw
        u(j1,j2)=0d0
30    continue
      do 35 j1=1,npw
        j2=jsym(j1)
        if (j2.eq.j1) then
          u(j1,j1)=1d0
        else if (j2.gt.j1) then
          u(j1,j1)=cos45
          u(j1,j2)=cos45
          u(j2,j1)=-cos45
          u(j2,j2)=cos45
        endif
35    continue
c
c     counts number of even and odd states:
c       for E-modes (E-field along z) even states are spatially even
c       for H-modes (H-field along z) even states are spatially odd
c
      neven=0
      nodd=0
      do 38 j1=1,npw
        j2=jsym(j1)
        if (j2.ge.j1) then
          if (jparxy.eq.1) neven=neven+1
          if (jparxy.eq.0) nodd=nodd+1
        else 
          if (jparxy.eq.1) nodd=nodd+1
          if (jparxy.eq.0) neven=neven++1
        endif
38    continue
      if (neven+nodd.ne.npw) stop ' from DECOUPLE: neven+nodd.ne.npw'
!      call wrmat (u,maxpw,npw,1d0,0d0)
c
c     transforms hamiltonian in the basis of even/odd states
c
      do 40 j1=1,npw
      do 40 j2=1,npw
!        work(j1,j2)=0d0
!        do 45 j3=1,npw
!          work(j1,j2)=work(j1,j2)+u(j1,j3)*h(j3,j2)
!45      continue
        j3=jsym(j1)
        if (j3.eq.j1) then
           work(j1,j2)=h(j1,j2)
        else
           work(j1,j2)=u(j1,j1)*h(j1,j2)+u(j1,j3)*h(j3,j2)
        endif
40    continue
      do 50 j1=1,npw
      do 50 j2=1,npw
!        h(j1,j2)=0d0
!        do 55 j3=1,npw
!          h(j1,j2)=h(j1,j2)+work(j1,j3)*u(j2,j3)
!55      continue
        j3=jsym(j2)
        if (j3.eq.j2) then
           h(j1,j2)=work(j1,j2)
        else
           h(j1,j2)=work(j1,j2)*u(j2,j2)+work(j1,j3)*u(j2,j3)
        endif
50    continue
      call repulisti (h,maxham,npw)
!      call wrmat (h,maxham,npw,1d0,0d0)
c
c     reorders blocks in order that even or odd states come first
c
      jeven=0
      jodd=0
      do 60 j1=1,npw
        j2=jsym(j1)
        if ((j2.ge.j1.and.jparxy.eq.1).or.
     &      (j2.lt.j1.and.jparxy.eq.0)) then ! index j1 belongs to even sector
          jeven=jeven+1
          jwork(j1)=jeven
          if (jparkz.eq.0) jwork(j1)=jeven
          if (jparkz.eq.1) jwork(j1)=nodd+jeven
        else if ((j2.ge.j1.and.jparxy.eq.0).or.
     &           (j2.lt.j1.and.jparxy.eq.1)) then ! j1 belongs to odd sector
          jodd=jodd+1
          if (jparkz.eq.0) jwork(j1)=neven+jodd
          if (jparkz.eq.1) jwork(j1)=jodd
        endif
60    continue
      do 62 j1=1,npw
      do 62 j2=1,npw
        work(jwork(j1),jwork(j2))=h(j1,j2)
62    continue
      do 65 j1=1,npw
      do 65 j2=1,npw
        h(j1,j2)=work(j1,j2)
65    continue
!      call wrmat (h,maxham,npw,1d0,0d0)
!      write (6,*) 'neven=',neven,'  nodd=',nodd
      if (jparkz.eq.0) nham=neven
      if (jparkz.eq.1) nham=nodd
      return
      end
c
      subroutine restore (eigenv,nham,nham_store)
c
c     restores eigenvectors in basis of plane waves
c
      implicit none
      integer maxpw,maxham,nham,nham_store,jham1,jham2,jham3
      parameter (maxpw=1000,maxham=2*maxpw)
      real*8 eigenv(maxham,maxham)
      real*8 u(maxham,maxham),work(maxham,maxham),sum
      integer jwork(maxham)
      common/matrix/u
      common/iarray/jwork
c
c     defines eigenvector component to be zero for all rows beyond nham
c
      do jham2=1,nham
        do jham1=nham+1,nham_store
          eigenv(jham1,jham2)=0d0
        enddo
      enddo
c
c     restores eigenvector in basis of even/odd combinations of PWs
c     (i.e., before using index jwork)
c
      do jham2=1,nham
        do jham1=1,nham_store
          work(jham1,jham2)=eigenv(jwork(jham1),jham2)
        enddo
      enddo
c
c     performs unitary transformation with inverse of matrix u
c
      do jham2=1,nham
        do jham1=1,nham_store
          sum=0d0
          do jham3=1,nham_store
            sum=sum+work(jham3,jham2)*u(jham3,jham1)
          enddo
          eigenv(jham1,jham2)=sum
        enddo
      enddo
      return
      end
c
      subroutine repulisti (a,max,n)
c
c     sets to zero elements smaller than UPPER
c
      implicit none
      integer max,n,j1,j2
      real*8 upper,a(max,max)
      parameter (upper=1d-12)
      do j1=1,n
      do j2=1,n
        if (abs(a(j1,j2)).lt.upper) a(j1,j2)=0d0
      enddo
      enddo
      return
      end
c
      subroutine checksym (a,max,n)
c
c     checks matrix A is symmetric
c
      implicit none
      integer max,n,j1,j2
      real*8 a(max,max),upper,diff
      parameter (upper=1d-12)
      if (n.gt.max) stop ' from CHECKSYM: n>max'
      do j1=1,n
      do j2=j1,n
        diff=a(j1,j2)-a(j2,j1)
        if (abs(diff).gt.upper) then
          write (6,*) ' from CHECKSYM: matrix A not symmetric'
          write (6,*) ' j1,j2,diff=',j1,j2,diff
        endif
      enddo
      enddo
      return
      end
c
      subroutine deblock (h,npw,block,j)
c
c     given Hamiltonian for jparxy=-1, defines the block
c     corresponding to the first or the second polarization vector
c
      implicit none
      integer j,maxpw,maxham,npw,j1,j2,jh1,jh2
      parameter (maxpw=1000,maxham=2*maxpw)
      real*8 h(maxham,maxham),block(maxham,maxham)
      do j1=1,npw
      do j2=1,npw
        if (j.eq.1) then
          jh1=2*j1
          jh2=2*j2
        else if (j.eq.2) then
          jh1=2*j1-1
          jh2=2*j2-1
        endif
        block(j1,j2)=h(jh1,jh2)
      enddo
      enddo
      return
      end
c
      double precision function spher0 (x)
c
c     spherical Bessel function of order 0: sin(x)/x
c
      implicit none
      real*8 x
      if (x.eq.0) then
        spher0=1d0
      else 
        spher0=sin(x)/x
      endif
      return
      end
c
      subroutine square_points (g,x,am)
c
c     main symmetry points of BZ of square lattice
c
      implicit none
      real*8 g(2),x(2),am(2),aret,pi
      parameter (aret=1d0)
      pi=4d0*atan(1d0)

      g(1)=0d0
      g(2)=0d0

      x(1)=pi/aret
      x(2)=0d0

      am(1)=pi/aret
      am(2)=pi/aret

      return
      end
c
      subroutine triang_points (g,am,ak)
c
c     main symmetry points of BZ of triangular lattice
c
      implicit none
      real*8 g(2),ak(2),am(2),aret,pi
      parameter (aret=1d0)
      pi=4d0*atan(1d0)

      g(1)=0d0
      g(2)=0d0

      am(1)=0d0
      am(2)=2d0*pi/sqrt(3d0)/aret

      ak(1)=2d0*pi/3d0/aret
      ak(2)=2d0*pi/sqrt(3d0)/aret

      return
      end
c
      subroutine w1_points (g,x1,x2,am)
c
c     main symmetry points of BZ of square lattice
c
      implicit none
      real*8 g(2),x1(2),x2(2),am(2),aret,pi
      real*8 alength1,alength2,alength3,dummy
      parameter (aret=1d0)
      common/length/alength1,alength2,alength3,dummy
      pi=4d0*atan(1d0)

      g(1)=0d0
      g(2)=0d0

      x1(1)=pi/aret
      x1(2)=0d0
      
      x2(1)=0d0
      x2(2)=pi/aret/(alength2+alength3)/sqrt(3d0)

      am(1)=pi/aret
      am(2)=pi/aret/(alength2+alength3)/sqrt(3d0)
      
      return
      end
c
      subroutine kline (p,q,akx,aky,ax,maxk,nk,jstart)
c
c     defines line in BZ from point p to q
c
      implicit none
      integer maxk,nk,jstart,jk
      real*8 p(2),q(2),akx(maxk),aky(maxk),ax(maxk),delta,dist
      
      if (jstart+nk.gt.maxk) stop ' from kline: jstart+nk>maxk'
      if (nk.ne.0) delta=1d0/nk
      dist=sqrt( (q(1)-p(1))*(q(1)-p(1)) + (q(2)-p(2))*(q(2)-p(2)) )
      do jk=0,nk
        akx(jstart+jk)=p(1)+(q(1)-p(1))*jk*delta
        aky(jstart+jk)=p(2)+(q(2)-p(2))*jk*delta
        ax (jstart+jk)=ax(jstart)+dist*jk*delta 
      enddo
      return
      end
c
      SUBROUTINE SORT5(N,RA,RB,RC,ND,NE,WKSP,IWKSP,nwksp)
      implicit real*8 (a-h,o-z)
      DIMENSION RA(N),RB(N),RC(N),ND(N),NE(N),WKSP(N),IWKSP(N),nwksp(n)
      CALL INDEXX(N,RA,IWKSP)
      DO 11 J=1,N
        WKSP(J)=RA(J)
11    CONTINUE
      DO 12 J=1,N
        RA(J)=WKSP(IWKSP(J))
12    CONTINUE
      DO 13 J=1,N
        WKSP(J)=RB(J)
13    CONTINUE
      DO 14 J=1,N
        RB(J)=WKSP(IWKSP(J))
14    CONTINUE
      DO 15 J=1,N
        WKSP(J)=RC(J)
15    CONTINUE
      DO 16 J=1,N
        RC(J)=WKSP(IWKSP(J))
16    CONTINUE
      DO 17 J=1,N
        NWKSP(J)=ND(j)
17    CONTINUE
      DO 18 J=1,N
        ND(j)=NWKSP(IWKSP(J))
18    CONTINUE
      DO 19 J=1,N
        NWKSP(J)=NE(J)
19    CONTINUE
      DO 20 J=1,N
        NE(J)=NWKSP(IWKSP(J))
20    CONTINUE
      RETURN
      END
c
c     *************************************************************
c
      SUBROUTINE INDEXX(N,ARRIN,INDX)
      implicit real*8 (a-h,o-z)
      DIMENSION ARRIN(N),INDX(N)
      DO 11 J=1,N
        INDX(J)=J
11    CONTINUE
      L=N/2+1
      IR=N
10    CONTINUE
        IF(L.GT.1)THEN
          L=L-1
          INDXT=INDX(L)
          Q=ARRIN(INDXT)
        ELSE
          INDXT=INDX(IR)
          Q=ARRIN(INDXT)
          INDX(IR)=INDX(1)
          IR=IR-1
          IF(IR.EQ.1)THEN
            INDX(1)=INDXT
            RETURN
          ENDIF
        ENDIF
        I=L
        J=L+L
20      IF(J.LE.IR)THEN
          IF(J.LT.IR)THEN
            IF(ARRIN(INDX(J)).LT.ARRIN(INDX(J+1)))J=J+1
          ENDIF
          IF(Q.LT.ARRIN(INDX(J)))THEN
            INDX(I)=INDX(J)
            I=J
            J=J+J
          ELSE
            J=IR+1
          ENDIF
        GO TO 20
        ENDIF
        INDX(I)=INDXT
      GO TO 10
      END
c
      subroutine savemt (a,b,nmax,n)
c
c     copies matrix A in matrix B
c

      implicit none
      integer nmax,n,j1,j2
      real*8 a(nmax,nmax),b(nmax,nmax)
      do j1=1,n
      do j2=1,n
        b(j1,j2)=a(j1,j2)
      enddo
      enddo
      return
      end
c
      double precision function delta_kronecker (j1,j2)
c
c     Kronecker's delta
c
      implicit none
      integer j1,j2
      if (j1.eq.j2) then
        delta_kronecker=1d0
      else
        delta_kronecker=0d0
      endif
      return
      end
c
      subroutine checkinv (a,b,nmax,n)
c
c     checks that matrices A and B are the inverse of each other,
c     with a tolerance given by parameter UPPER
c
      implicit none
      integer nmax,n,j1,j2,j3
      real*8 a(nmax,nmax),b(nmax,nmax),sum,upper
      parameter (upper=1d-12)
      do j1=1,n
      do j2=1,n
        sum=0d0
        do j3=1,n
          sum=sum+a(j1,j3)*b(j3,j2)
        enddo
        if (j1.eq.j2) then
          if (abs(sum-1d0).gt.upper) stop ' from checkinv: error'
        else
          if (abs(sum).gt.upper) stop ' from checkinv: error'
        endif
      enddo
      enddo
      return
      end
c
      double precision function scal(a,b,n)
c
c     scalar product of arrays
c
      implicit none
      integer n,j
      real*8 a(n),b(n)
      scal=0d0
      do j=1,n
        scal=scal+a(j)*b(j)
      enddo
      return
      end
c
      FUNCTION BESSJ1(X)
c
c     Bessel function of order 1: from Numerical Recipes
c
      implicit real*8 (a-h,o-z)
      REAL*8 Y,P1,P2,P3,P4,P5,Q1,Q2,Q3,Q4,Q5,R1,R2,R3,R4,R5,R6,
     *    S1,S2,S3,S4,S5,S6
      DATA R1,R2,R3,R4,R5,R6/72362614232.D0,-7895059235.D0,242396853.1D0
     *,
     *    -2972611.439D0,15704.48260D0,-30.16036606D0/,
     *    S1,S2,S3,S4,S5,S6/144725228442.D0,2300535178.D0,
     *    18583304.74D0,99447.43394D0,376.9991397D0,1.D0/
      DATA P1,P2,P3,P4,P5/1.D0,.183105D-2,-.3516396496D-4,.2457520174D-5
     *,
     *    -.240337019D-6/, Q1,Q2,Q3,Q4,Q5/.04687499995D0,-.2002690873D-3
     *,
     *    .8449199096D-5,-.88228987D-6,.105787412D-6/
      IF(ABS(X).LT.8.)THEN
        Y=X**2
        BESSJ1=X*(R1+Y*(R2+Y*(R3+Y*(R4+Y*(R5+Y*R6)))))
     *      /(S1+Y*(S2+Y*(S3+Y*(S4+Y*(S5+Y*S6)))))
      ELSE
        AX=ABS(X)
        Z=8.d0/AX
        Y=Z**2
        XX=AX-2.356194491
        BESSJ1=SQRT(.636619772/AX)*(COS(XX)*(P1+Y*(P2+Y*(P3+Y*(P4+Y
     *      *P5))))-Z*SIN(XX)*(Q1+Y*(Q2+Y*(Q3+Y*(Q4+Y*Q5)))))
     *      *SIGN(1.d0,X)
      ENDIF
      RETURN
      END
c
      SUBROUTINE WRMAT (A,NMAX,N,FACPRO,FACSUM)
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 A(NMAX,NMAX)
      WRITE (6,*)
C
C     NUMBER OF TIMES MUST WRITE BLOCKS OF COLUMNS
C
      NWRITE=1+(N-1)/10
      DO 30 IWRITE=1,NWRITE
C
C       N1,..,N2=NUMBER OF COLUMNS MUST BE WRITTEN IN THE PRESENT BLOCK
C
        N1=1+10*(IWRITE-1)
        N2=MIN(10*IWRITE,N)
        IF (NWRITE.GT.1) WRITE (6,5) N1,N2
5       FORMAT(' COLUMNS ',I2,' - ',I2,/)
        DO 10 I=1,N
10        WRITE (6,20) ((A(I,J)+FACSUM)*FACPRO,J=N1,N2)
20        FORMAT(1X,10(E10.4,2X))
        WRITE (6,*)
30    CONTINUE
      WRITE (6,*)
      RETURN
      END
C
      SUBROUTINE WRVEC (A,NMAX,N,FACPRO,FACSUM)
      IMPLICIT REAL*8 (A-H,O-Z)
      REAL*8 A(NMAX)
      WRITE (6,*)
      NWRITE=1+(N-1)/10
      DO 30 IWRITE=1,NWRITE
        N1=1+10*(IWRITE-1)
        N2=MIN(10*IWRITE,N)
        WRITE (6,20) ((A(J)+FACSUM)*FACPRO,J=N1,N2)
20      FORMAT(1X,10(E10.4,2X))
        WRITE (6,*)
30    CONTINUE
      RETURN
      END
C
      SUBROUTINE WRMAT2 (A,NMAX,NBLOCK,NDIM,JBLOCK,FACPRO,FACSUM)
      IMPLICIT REAL*8 (A-H,O-Z)
      PARAMETER (N2=40)
      REAL*8 A(NMAX,NMAX),BLOCK(N2,N2)
      IF (NDIM.GT.N2) THEN
        WRITE (6,2) N2
2       FORMAT(' FROM MODULE WRMAT2: BLOCK DIMENSION EXCEEDS NMAX=',I2)
        WRITE (6,*) ' NO MATRIX WRITTEN'
        RETURN
      ENDIF
C
C     DEFINES BLOCK TO BE WRITTEN
C
      DO 10 I=1,NDIM
       DO 10 J=1,NDIM
10      BLOCK(I,J)=A((JBLOCK-1)*NDIM+I,(JBLOCK-1)*NDIM+J)
      CALL WRMAT (BLOCK,N2,NDIM,FACPRO,FACSUM)
      RETURN
      END
C
      SUBROUTINE WRVEC2 (A,NMAX,NBLOCK,NDIM,JBLOCK,FACPRO,FACSUM)
      IMPLICIT REAL*8 (A-H,O-Z)
      PARAMETER (N2=40)
      REAL*8 A(NMAX),BLOCK(N2)
      IF (NDIM.GT.N2) THEN
        WRITE (6,2) N2
2       FORMAT(' FROM MODULE WRVEC2: BLOCK DIMENSION EXCEEDS NMAX=',I2)
        WRITE (6,*) ' NO VECTOR WRITTEN'
        RETURN
      ENDIF
C
C     DEFINES BLOCK TO BE WRITTEN
C
      DO 10 I=1,NDIM
10      BLOCK(I)=A((JBLOCK-1)*NDIM+I)
      CALL WRVEC (BLOCK,N2,NDIM,FACPRO,FACSUM)
      RETURN
      END
